from datetime import datetime, timedelta
from time import monotonic
from typing import Union

__all__ = ["monotonic_time_from_deadline"]


def monotonic_time_from_deadline(deadline: Union[datetime, timedelta]) -> float:
    if isinstance(deadline, timedelta):
        duration = deadline.total_seconds()
    else:
        duration = (deadline - datetime.now()).total_seconds()

    return monotonic() + duration



from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, TypedDict

import grpc
from aserto.directory.common.v2 import (
    Object,
    ObjectIdentifier,
    ObjectTypeIdentifier,
    PaginationRequest,
    PaginationResponse,
    PermissionIdentifier,
    Relation,
    RelationIdentifier,
    RelationTypeIdentifier,
)
from aserto.directory.exporter.v2 import ExporterStub
from aserto.directory.importer.v2 import ImporterStub
from aserto.directory.reader.v2 import (
    CheckPermissionRequest,
    CheckRelationRequest,
    GetObjectManyRequest,
    GetObjectRequest,
    GetObjectsRequest,
    GetObjectsResponse,
    GetRelationRequest,
    GetRelationsRequest,
    GetRelationsResponse,
    ReaderStub,
)
from aserto.directory.writer.v2 import (
    DeleteObjectRequest,
    DeleteRelationRequest,
    SetObjectRequest,
    SetRelationRequest,
    WriterStub,
)


class NotFoundError(Exception):
    pass


class Directory:
    def __init__(self, channel: grpc.Channel, api_key: str, tenant_id: str) -> None:
        self._channel = channel
        self._metadata = self._get_metadata(api_key=api_key, tenant_id=tenant_id)
        self.reader = ReaderStub(self._channel)
        self.writer = WriterStub(self._channel)
        self.importer = ImporterStub(self._channel)
        self.exporter = ExporterStub(self._channel)

    @classmethod
    def connect(cls, *, address: str, api_key: str, tenant_id: str, ca_cert: str):
        channel = grpc.secure_channel(
            target=address, credentials=cls._channel_credentials(cert=ca_cert)
        )
        return Directory(channel, api_key, tenant_id)

    def get_objects(
        self, object_type: Optional[str] = None, page: Optional[PaginationRequest] = None
    ) -> GetObjectsResponse:
        """Retrieve a page of directory objects by object type and page size.
        Returns a list of directory objects and a token for the next page if it exists.

        Parameters
        ----
        object_type : str
            a directory object type
        page : PaginationRequest(size: int, token: str)
            paging information — the size of the page, and the pagination
            start token

        Returns
        ----
        GetObjectsResponse
            results : list(Object)
                list of directory objects
            page : PaginationResponse(result_size: int, next_token: str)
                retrieved page information — the size of the page,
                and the next page's token
        """

        response = self.reader.GetObjects(
            GetObjectsRequest(param=ObjectTypeIdentifier(name=object_type), page=page),
            metadata=self._metadata,
        )
        return response

    def get_many_objects(
        self,
        objects: Optional[List[TypedDict("ObjectParams", {"key": str, "type": str})]] = None,
    ) -> List[Object]:
        """Retrieve a list of directory object using a list of object key and type pairs.
        Returns a list of each objects, if an object with the specified key and type exists.

        Parameters
        ----
        objects : list( dict(key: str, type: str) )
            list of object key and object type pairs

        Returns
        ----
        list
            list of directory objects
        """

        identifiers = [ObjectIdentifier(key=x["key"], type=x["type"]) for x in objects]
        print("identifiers", identifiers)
        response = self.reader.GetObjectMany(
            GetObjectManyRequest(param=identifiers),
            metadata=self._metadata,
        )
        return response.results

    def get_object(self, key: str, type: str) -> Object:
        """Retrieve a directory object by its key and type.
        Returns the object or raises a NotFoundError if an object with the
        specified key and type doesn't exist.

        Parameters
        ----
        key : str
            an object key
        type : str
            an object type

        Returns
        ----
        a directory object
        """

        try:
            identifier = ObjectIdentifier(type=type, key=key)
            response = self.reader.GetObject(
                GetObjectRequest(param=identifier), metadata=self._metadata
            )
            return response.result

        except grpc.RpcError as err:
            if err.code() == grpc.StatusCode.NOT_FOUND:
                raise NotFoundError from err
            raise

    def set_object(self, object: Object) -> Object:
        """Updates a directory object given its key and type, or creates a new object.
        Returns the created/updated object.

        Parameters
        ----
        object : Object
            key : str
            type : str
            display_name : str, optional
            properties : dict, optional
            hash : str (required, if updating existing object)

        Returns
        ----
        a directory object
        """

        response = self.writer.SetObject(SetObjectRequest(object=object), metadata=self._metadata)
        return response.result

    def delete_object(self, key: str, type: str) -> None:
        """Deletes a directory object given its key and type.
        Returns None.

        Parameters
        ----
        key : str
            an object key
        type : str
            an object type

        Returns
        ----
        None
        """

        identifier = ObjectIdentifier(type=type, key=key)
        self.writer.DeleteObject(DeleteObjectRequest(param=identifier), metadata=self._metadata)

    def get_relations(
        self,
        subject_type: Optional[str] = None,
        subject_key: Optional[str] = None,
        object_type: Optional[str] = None,
        object_key: Optional[str] = None,
        relation_type: Optional[str] = None,
        page: Optional[PaginationRequest] = None,
    ) -> GetRelationsResponse:
        """Retrieve a page of directory relations by any of the parameters
        object's type and key, subject's type and key, relation type name, and page size.
        Returns a list of directory relations and a token for the next page if it exists.
        object --relation_type--> subject

        Parameters
        ----
        subject_type : str
            the subject of the relation, a directory object type
        subject_key : str
            the subject of the relation, a directory object key
        object_type : str
            the object of the relation, a directory object type
        object_key : str
            the object of the relation, a directory object key
        relation_type : str
            a directory relation type
        page : PaginationRequest(size: int, token: str)
            paging information — the size of the page, and the pagination
            start token

        Returns
        ----
        GetRelationsResponse
            results : list(Relation)
                list of directory relations
            page : PaginationResponse(result_size: int, next_token: str)
                retrieved page information — the size of the page,
                and the next page's token
        """

        response = self.reader.GetRelations(
            GetRelationsRequest(
                param=RelationIdentifier(
                    object=ObjectIdentifier(type=object_type, key=object_key),
                    subject=ObjectIdentifier(type=subject_type, key=subject_key),
                    relation=RelationTypeIdentifier(name=relation_type, object_type=object_type),
                ),
                page=page,
            ),
            metadata=self._metadata,
        )
        return response

    def get_relation(
        self,
        subject_type: Optional[str] = None,
        subject_key: Optional[str] = None,
        object_type: Optional[str] = None,
        object_key: Optional[str] = None,
        relation_type: Optional[str] = None,
        with_objects: Optional[bool] = None,
    ) -> Dict[Relation, Optional[Dict[str, Object]]]:
        """Retrieve a directory relation by the object's type and key, the subject's type and key,
        and relation type name.
        Returns the relation or raises a NotFoundError if an relation with the
        specified parameters doesn't exist. Also returns the object if with_objects is set to True.
        object --relation_type--> subject

        Parameters
        ----
        subject_type : str
            the subject of the relation, a directory object type
        subject_key : str
            the subject of the relation, a directory object key
        object_type : str (required if relation_type is specified)
            the object of the relation, a directory object type
        object_key : str
            the object of the relation, a directory object key
        relation_type : str
            a directory relation type
        with_objects : bool
            if True, returns the object

        Returns
        ----
        relation : Relation
            a directory relations
        objects : dict(str, Object)
            returned with an Object if with_objects is set to True
        """

        response = self.reader.GetRelation(
            GetRelationRequest(
                param=RelationIdentifier(
                    object=ObjectIdentifier(type=object_type, key=object_key),
                    subject=ObjectIdentifier(type=subject_type, key=subject_key),
                    relation=RelationTypeIdentifier(name=relation_type, object_type=object_type),
                ),
                with_objects=with_objects,
            ),
            metadata=self._metadata,
        )

        if not len(response.results):
            raise NotFoundError
        return {"relation": response.results[0], "objects": response.objects}

    def set_relation(self, relation: Relation) -> Relation:
        """Updates a directory relation given the relation name and object type,
        or creates a new relation.
        Returns the created/updated object.

        Parameters
        ----
        relation : Relation
            subject: Object
            relation: str
            object: Object
            hash: str (required, if updating existing relation)

        Returns
        ----
        a directory relation
        """

        response = self.writer.SetRelation(
            SetRelationRequest(relation=relation), metadata=self._metadata
        )
        return response.result

    def delete_relation(
        self,
        subject_type: str,
        subject_key: str,
        object_type: str,
        object_key: str,
        relation_type: str,
    ) -> None:
        """Deletes a directory relation the object's type and key, the subject's type and key,
        and relation type name.
        Returns None.
        object --relation_type--> subject

        Parameters
        ----
        subject_type : str
            the subject of the relation, a directory object type
        subject_key : str
            the subject of the relation, a directory object key
        object_type : str
            the object of the relation, a directory object type
        object_key : str
            the object of the relation, a directory object key
        relation_type : str
            a directory relation type

        Returns
        ----
        None
        """

        relation_identifier = RelationIdentifier(
            object=ObjectIdentifier(type=object_type, key=object_key),
            subject=ObjectIdentifier(type=subject_type, key=subject_key),
            relation=RelationTypeIdentifier(name=relation_type, object_type=object_type),
        )
        self.writer.DeleteRelation(
            DeleteRelationRequest(param=relation_identifier), metadata=self._metadata
        )

    def check_relation(
        self,
        subject_type: str,
        subject_key: str,
        object_type: str,
        object_key: str,
        relation_type: str,
    ) -> bool:
        """Checks if a subject has a given relation to an object given
        the object's type and key, the subject's type and key,
        and relation type name.
        Returns the result of the relation check, True or False.
        object --relation_type--> subject

        Parameters
        ----
        subject_type : str
            the subject of the relation, a directory object type
        subject_key : str
            the subject of the relation, a directory object key
        object_type : str
            the object of the relation, a directory object type
        object_key : str
            the object of the relation, a directory object key
        relation_type : str
            a directory relation type

        Returns
        ----
        True or False
        """

        response = self.reader.CheckRelation(
            CheckRelationRequest(
                object=ObjectIdentifier(type=object_type, key=object_key),
                subject=ObjectIdentifier(type=subject_type, key=subject_key),
                relation=RelationTypeIdentifier(name=relation_type, object_type=object_type),
            ),
            metadata=self._metadata,
        )
        return response.check

    def check_permission(
        self,
        subject_type: str,
        subject_key: str,
        object_type: str,
        object_key: str,
        permission: str,
    ) -> bool:
        """Checks if a subject has a given permission on an object given
        the object's type and key, the subject's type and key,
        and permission name.
        Returns the result of the permission check, True or False.

        Parameters
        ----
        subject_type : str
            the subject to check the permission for, a directory object type
        subject_key : str
            the subject to check the permission for, a directory object key
        object_type : str
            the object to check the permission on, a directory object type
        object_key : str
            the object to check the permission on, a directory object key
        permission : str
            a directory permission

        Returns
        ----
        True or False
        """

        response = self.reader.CheckPermission(
            CheckPermissionRequest(
                object=ObjectIdentifier(type=object_type, key=object_key),
                subject=ObjectIdentifier(type=subject_type, key=subject_key),
                permission=PermissionIdentifier(name=permission),
            ),
            metadata=self._metadata,
        )
        return response.check

    def close(self) -> None:
        """Closes the gRPC channel"""

        self._channel.close()

    @staticmethod
    def _get_metadata(api_key, tenant_id) -> Tuple:
        md = ()
        if api_key:
            md += (("authorization", f"basic {api_key}"),)
        if tenant_id:
            md += (("aserto-tenant-id", tenant_id),)
        return md

    @staticmethod
    def _channel_credentials(cert) -> grpc.ChannelCredentials:
        if cert:
            with open(cert, "rb") as f:
                return grpc.ssl_channel_credentials(f.read())
        else:
            return grpc.ssl_channel_credentials()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



import ssl
from typing import Mapping, Optional

from typing_extensions import Literal

from ._typing import assert_unreachable

__all__ = ["AuthorizerOptions"]


ASERTO_HOSTED_AUTHORIZER_URL = "https://authorizer.prod.aserto.com"


ServiceType = Literal["gRPC", "REST"]


class AuthorizerOptions:
    def __init__(
        self,
        *,
        url: str = ASERTO_HOSTED_AUTHORIZER_URL,
        tenant_id: Optional[str] = None,
        api_key: Optional[str] = None,
        cert_file_path: Optional[str] = None,
        service_type: ServiceType = "gRPC",
    ):
        self._tenant_id = tenant_id
        self._api_key = api_key
        self._cert_file_path = cert_file_path
        self._service_type = service_type

        if not url.endswith("aserto.com"):
            self._url = url
        elif service_type == "gRPC":
            self._url = f"{url}:8443"
        elif service_type == "REST":
            self._url = url
        else:
            assert_unreachable(service_type)

    @property
    def url(self) -> str:
        return self._url

    @property
    def api_key(self) -> Optional[str]:
        return self._api_key

    @property
    def tenant_id(self) -> Optional[str]:
        return self._tenant_id

    @property
    def service_type(self) -> ServiceType:
        return self._service_type

    @property
    def cert(self) -> Optional[bytes]:
        if self._cert_file_path is None:
            return None

        with open(self._cert_file_path, "rb") as f:
            return f.read()

    @property
    def ssl_context(self) -> Optional[ssl.SSLContext]:
        return ssl.create_default_context(cafile=self._cert_file_path) if self._cert_file_path else None

    @property
    def auth_headers(self) -> Mapping[str, str]:
        headers = {}
        if self._api_key:
            headers["authorization"] = f"basic {self._api_key}"
        if self._tenant_id:
            headers["aserto-tenant-id"] = self._tenant_id

        return headers



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aserto',
 'aserto.client',
 'aserto.client.api',
 'aserto.client.api.authorizer',
 'aserto.client.directory']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'aserto-authorizer==0.20.0',
 'aserto-directory',
 'grpcio>=1.49.1,<2.0.0',
 'protobuf>=4.21.7,<5.0.0',
 'typing-extensions>=3.10.0,<4.0.0']

setup_kwargs = {
    'name': 'aserto',
    'version': '0.21.0',
    'description': 'Aserto API client',
    'long_description': '# Aserto API client\n\nHigh-level client interface to Aserto\'s APIs.\n\nAt the moment this only supports interacting with Aserto\'s [Authorizer service](https://docs.aserto.com/docs/authorizer-guide/overview).\n\n## Installation\n\n### Using Pip\n\n```sh\npip install aserto\n```\n\n### Using Poetry\n\n```sh\npoetry add aserto\n```\n\n## Usage\n\n```py\nfrom aserto.client import AuthorizerOptions, Identity\nfrom aserto.client.api.authorizer import AuthorizerClient\n\n\nclient = AuthorizerClient(\n    identity=Identity(type="NONE"),\n    options=AuthorizerOptions(\n        api_key=ASERTO_API_KEY,\n        tenant_id=ASERTO_TENANT_ID,\n        service_type="gRPC",\n    ),\n)\n\nresult = await client.decision_tree(\n    decisions=["visible", "enabled", "allowed"],\n    policy_instance_name=ASERTO_POLICY_INSTANCE_NAME,\n    policy_instance_label=ASERTO_POLICY_INSTANCE_LABEL,\n    policy_path_root=ASERTO_POLICY_PATH_ROOT,\n    policy_path_separator="DOT",\n)\n\nassert result == {\n    "GET.your.policy.path": {\n        "visible": True,\n        "enabled": True,\n        "allowed": False,\n    },\n}\n```\n\n## Directory\n\nThe Directory APIs can be used to get or set object instances and relation instances. They can also be used to check whether a user has a permission or relation on an object instance.\n\n### Directory Client\n\nYou can initialize a directory client as follows:\n\n```py\nfrom aserto.client.directory import Directory\n\nds = Directory.connect(api_key="my_api_key", tenant_id="1234", address="localhost:9292")\n```\n\n- `address`: hostname:port of directory service (_required_)\n- `api_key`: API key for directory service (_required_ if using hosted directory)\n- `tenant_id`: Aserto tenant ID (_required_ if using hosted directory)\n- `cert`: Path to the grpc service certificate when connecting to local topaz instance.\n\n#### \'get_object\' function\n\nGet a directory object instance with the type and the key.\n\n```py\nuser = ds.get_object(type="user", key="euang@acmecorp.com")\n```\n\n#### \'get_objects\' function\n\nGet object instances with an object type type and page size.\n\n```py\nfrom aserto.client.directory import PaginationRequest\n\nusers = ds.get_objects(object_type="user", page=PaginationRequest(size=10))\n```\n\n#### \'set_object\' function\n\nCreate an object instance with the specified fields. For example:\n\n```py\nfrom google.protobuf.json_format import ParseDict\nfrom google.protobuf.struct_pb2 import Struct\n\nproperties = ParseDict({"displayName": "test object"}, Struct())\n\nuser = ds.set_object(object={\n    "type": "user",\n    "key": "test-object",\n    "properties": properties,\n})\n```\n\n#### \'delete_object\' function\n\nDelete an object instance using its type and key:\n\n```py\nds.delete_object(type="user", key="test-object")\n```\n\n## License\n\nThis project is licensed under the MIT license. See the [LICENSE](https://github.com/aserto-dev/aserto-python/blob/main/LICENSE) file for more info.\n',
    'author': 'Aserto, Inc.',
    'author_email': 'pypi@aserto.com',
    'maintainer': 'authereal',
    'maintainer_email': 'authereal@aserto.com',
    'url': 'https://github.com/aserto-dev/aserto-python/tree/HEAD/packages/aserto',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
